{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Spectrum analysis\n",
    "\n",
    "\n",
    "For the analysis of obtained spectra we will focus on the carbon X-ray absorption spectrum of vinylfluoride, with emphasis on how the obtained spectra can be decomposed in terms of the polarization of incoming radiation, and the atomic origin of the features. Additional analysis in terms of MOs, different [visualization](../visualize/exc_vis) schemes and descriptors thereof, will be added at a later date, as well as more details on assignment. Analysis of features in X-ray emission spectra tends to be easier to interpret (fewer states, probing occupied states of relatively large energy shifts, etc.), and focus will thus here be on XAS.\n",
    "\n",
    "Vinylfluride (C$_2$H$_3$F) is a fluorine-substituted ethene derivate, which possess large shifts in features for the -**C**HF site due to interaction with the very electronegative fluorine. This *chemical shift* is significant and a strong indicator of the local chemical environment. This can be seen by comparing the experimental spectra of ethene, vinylfluoride, and 1,1-difluoroethene, as digitized from [experimental measurements](https://doi.org/10.1103/PhysRevA.36.1683). We here mark position (and number of contributing carbon sites) of strong $1s \\rightarrow \\pi^{\\ast}$ transitions.\n",
    "\n",
    "\n",
    "```{figure} ../../img/xray/expt_ethene.svg\n",
    "---\n",
    "name: /expt_ethene\n",
    "---\n",
    "```\n",
    "\n",
    "We see that a single substitution yields a shift of about 2 eV, and a double substitution (at the same site) of about 4 eV. The behaviour of the chemical shifts for different spectroscopies will be further deliberated on in [exercises](sec:xray_ex_chem)."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "Calculating the vinylfluoride spectrum using CVS-ADC(2)-x and CPP-DFT (with the B3LYP xc-functional):\n",
    "\n",
    "```python\n",
    "vinyl_mol_str = \"\"\"\n",
    " C     0.000000    -0.246412    -1.271068\n",
    " C     0.000000     0.457081    -0.154735\n",
    " F     0.000000    -0.119195     1.052878\n",
    " H     0.000000     0.272328    -2.210194\n",
    " H     0.000000    -1.319906    -1.249847\n",
    " H     0.000000     1.530323    -0.095954\n",
    "\"\"\"\n",
    "\n",
    "# Construct structure and basis objects\n",
    "struct = gator.get_molecule(vinyl_mol_str)\n",
    "basis = gator.get_molecular_basis(struct, \"6-31G\")\n",
    "```"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {
    "tags": [
     "remove-input"
    ]
   },
   "outputs": [
    {
     "data": {
      "application/3dmoljs_load.v0": "<div id=\"3dmolviewer_16794854282110507\"  style=\"position: relative; width: 300px; height: 200px\">\n        <p id=\"3dmolwarning_16794854282110507\" style=\"background-color:#ffcccc;color:black\">You appear to be running in JupyterLab (or JavaScript failed to load for some other reason).  You need to install the 3dmol extension: <br>\n        <tt>jupyter labextension install jupyterlab_3dmol</tt></p>\n        </div>\n<script>\n\nvar loadScriptAsync = function(uri){\n  return new Promise((resolve, reject) => {\n    //this is to ignore the existence of requirejs amd\n    var savedexports, savedmodule;\n    if (typeof exports !== 'undefined') savedexports = exports;\n    else exports = {}\n    if (typeof module !== 'undefined') savedmodule = module;\n    else module = {}\n\n    var tag = document.createElement('script');\n    tag.src = uri;\n    tag.async = true;\n    tag.onload = () => {\n        exports = savedexports;\n        module = savedmodule;\n        resolve();\n    };\n  var firstScriptTag = document.getElementsByTagName('script')[0];\n  firstScriptTag.parentNode.insertBefore(tag, firstScriptTag);\n});\n};\n\nif(typeof $3Dmolpromise === 'undefined') {\n$3Dmolpromise = null;\n  $3Dmolpromise = loadScriptAsync('https://cdnjs.cloudflare.com/ajax/libs/3Dmol/2.0.1/3Dmol-min.js');\n}\n\nvar viewer_16794854282110507 = null;\nvar warn = document.getElementById(\"3dmolwarning_16794854282110507\");\nif(warn) {\n    warn.parentNode.removeChild(warn);\n}\n$3Dmolpromise.then(function() {\nviewer_16794854282110507 = $3Dmol.createViewer(document.getElementById(\"3dmolviewer_16794854282110507\"),{backgroundColor:\"white\"});\nviewer_16794854282110507.zoomTo();\n\tviewer_16794854282110507.addModel(\"6\\n\\n C     0.000000    -0.246412    -1.271068\\n C     0.000000     0.457081    -0.154735\\n F     0.000000    -0.119195     1.052878\\n H     0.000000     0.272328    -2.210194\\n H     0.000000    -1.319906    -1.249847\\n H     0.000000     1.530323    -0.095954\\n\",\"xyz\");\n\tviewer_16794854282110507.setViewStyle({\"style\": \"outline\", \"width\": 0.05});\n\tviewer_16794854282110507.setStyle({\"stick\": {}, \"sphere\": {\"scale\": 0.25}});\n\tviewer_16794854282110507.rotate(90,\"y\");\nviewer_16794854282110507.render();\n});\n</script>",
      "text/html": [
       "<div id=\"3dmolviewer_16794854282110507\"  style=\"position: relative; width: 300px; height: 200px\">\n",
       "        <p id=\"3dmolwarning_16794854282110507\" style=\"background-color:#ffcccc;color:black\">You appear to be running in JupyterLab (or JavaScript failed to load for some other reason).  You need to install the 3dmol extension: <br>\n",
       "        <tt>jupyter labextension install jupyterlab_3dmol</tt></p>\n",
       "        </div>\n",
       "<script>\n",
       "\n",
       "var loadScriptAsync = function(uri){\n",
       "  return new Promise((resolve, reject) => {\n",
       "    //this is to ignore the existence of requirejs amd\n",
       "    var savedexports, savedmodule;\n",
       "    if (typeof exports !== 'undefined') savedexports = exports;\n",
       "    else exports = {}\n",
       "    if (typeof module !== 'undefined') savedmodule = module;\n",
       "    else module = {}\n",
       "\n",
       "    var tag = document.createElement('script');\n",
       "    tag.src = uri;\n",
       "    tag.async = true;\n",
       "    tag.onload = () => {\n",
       "        exports = savedexports;\n",
       "        module = savedmodule;\n",
       "        resolve();\n",
       "    };\n",
       "  var firstScriptTag = document.getElementsByTagName('script')[0];\n",
       "  firstScriptTag.parentNode.insertBefore(tag, firstScriptTag);\n",
       "});\n",
       "};\n",
       "\n",
       "if(typeof $3Dmolpromise === 'undefined') {\n",
       "$3Dmolpromise = null;\n",
       "  $3Dmolpromise = loadScriptAsync('https://cdnjs.cloudflare.com/ajax/libs/3Dmol/2.0.1/3Dmol-min.js');\n",
       "}\n",
       "\n",
       "var viewer_16794854282110507 = null;\n",
       "var warn = document.getElementById(\"3dmolwarning_16794854282110507\");\n",
       "if(warn) {\n",
       "    warn.parentNode.removeChild(warn);\n",
       "}\n",
       "$3Dmolpromise.then(function() {\n",
       "viewer_16794854282110507 = $3Dmol.createViewer(document.getElementById(\"3dmolviewer_16794854282110507\"),{backgroundColor:\"white\"});\n",
       "viewer_16794854282110507.zoomTo();\n",
       "\tviewer_16794854282110507.addModel(\"6\\n\\n C     0.000000    -0.246412    -1.271068\\n C     0.000000     0.457081    -0.154735\\n F     0.000000    -0.119195     1.052878\\n H     0.000000     0.272328    -2.210194\\n H     0.000000    -1.319906    -1.249847\\n H     0.000000     1.530323    -0.095954\\n\",\"xyz\");\n",
       "\tviewer_16794854282110507.setViewStyle({\"style\": \"outline\", \"width\": 0.05});\n",
       "\tviewer_16794854282110507.setStyle({\"stick\": {}, \"sphere\": {\"scale\": 0.25}});\n",
       "\tviewer_16794854282110507.rotate(90,\"y\");\n",
       "viewer_16794854282110507.render();\n",
       "});\n",
       "</script>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "import py3Dmol as p3d\n",
    "\n",
    "vinyl_xyz = \"\"\"6\n",
    "\n",
    " C     0.000000    -0.246412    -1.271068\n",
    " C     0.000000     0.457081    -0.154735\n",
    " F     0.000000    -0.119195     1.052878\n",
    " H     0.000000     0.272328    -2.210194\n",
    " H     0.000000    -1.319906    -1.249847\n",
    " H     0.000000     1.530323    -0.095954\n",
    "\"\"\"\n",
    "\n",
    "viewer = p3d.view(width=300, height=200)\n",
    "viewer.addModel(vinyl_xyz, 'xyz')\n",
    "viewer.setViewStyle({\"style\": \"outline\", \"width\": 0.05})\n",
    "viewer.setStyle({\"stick\":{},\"sphere\": {\"scale\":0.25}})\n",
    "viewer.rotate(90,'y')\n",
    "viewer.show()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "```python\n",
    "# Perform SCF calculation\n",
    "scf_gs = gator.run_scf(struct, basis)\n",
    "\n",
    "# Calculate the 10 lowest eigenstates with CVS restriction to MOs #1-#3 (fluorine and carbon 1s)\n",
    "adc_xas = gator.run_adc(\n",
    "    struct, basis, scf_gs, method=\"cvs-adc2x\", singlets=10, core_orbitals=[1, 2, 3]\n",
    ")\n",
    "```\n",
    "\n",
    "```{note}\n",
    "If you have a recent version of adcc, there may be convergence issues due to a mismatch between conda and adcc. This is being looked into, and can for the moment typically be avoided by increasing the basis set size.\n",
    "```\n",
    "\n",
    "```python\n",
    "# Prepare molecule and basis objects\n",
    "molecule = vlx.Molecule.read_molecule_string(vinyl_mol_str)\n",
    "basis = vlx.MolecularBasis.read(molecule, \"6-31G\")\n",
    "\n",
    "# SCF settings and calculation\n",
    "scf_drv = vlx.ScfRestrictedDriver()\n",
    "scf_settings = {\"conv_thresh\": 1.0e-6}\n",
    "method_settings = {\"xcfun\": \"b3lyp\"}\n",
    "scf_drv.update_settings(scf_settings, method_settings)\n",
    "scf_results = scf_drv.compute(molecule, basis)\n",
    "\n",
    "# Define spectrum region to be resolved\n",
    "freqs = np.arange(272.0, 285.0, 0.25) / au2ev\n",
    "freqs_str = [str(x) for x in freqs]\n",
    "\n",
    "# Calculate the response\n",
    "cpp_prop = LinearAbsorptionCrossSection(\n",
    "    {\"frequencies\": \",\".join(freqs_str), \"damping\": 0.4 / au2ev}, method_settings\n",
    ")\n",
    "cpp_prop.init_driver()\n",
    "cpp_prop.compute(molecule, basis, scf_results)\n",
    "\n",
    "# Extract the imaginary part of the complex response function and convert to absorption cross section\n",
    "sigma = []\n",
    "for w in freqs:\n",
    "    axx = -cpp_prop.rsp_property[\"response_functions\"][(\"x\", \"x\", w)].imag\n",
    "    ayy = -cpp_prop.rsp_property[\"response_functions\"][(\"y\", \"y\", w)].imag\n",
    "    azz = -cpp_prop.rsp_property[\"response_functions\"][(\"z\", \"z\", w)].imag\n",
    "    alpha_bar = (axx + ayy + azz) / 3.0\n",
    "    sigma.append(4.0 * np.pi * w * alpha_bar / 137.035999)\n",
    "```\n",
    "\n",
    "The resulting spectra are shown below, and look very similar (although with a large difference in absolute energy):\n",
    "\n",
    "```python\n",
    "plt.figure(figsize=(8, 5))\n",
    "# CPP-DFT results\n",
    "plt.subplot(211)\n",
    "plt.title(\"CPP-DFT (B3LYP)\")\n",
    "x = np.arange(min(au2ev * freqs), max(au2ev * freqs), 0.01)\n",
    "y = interp1d(au2ev * freqs, sigma, kind=\"cubic\")\n",
    "plt.plot(x, y(x), \"b\")\n",
    "plt.xlim(272, 284)\n",
    "\n",
    "# CVS-ADC(2)-x results\n",
    "plt.subplot(212)\n",
    "plt.title(\"CVS-ADC(2)-x\")\n",
    "x, y = au2ev * adc_xas.excitation_energy, adc_xas.oscillator_strength\n",
    "xi, yi = lorentzian(x, y, 280, 300, 0.01, 0.8)\n",
    "plt.plot(xi, yi, \"b\")\n",
    "plt.xlim(284, 296)\n",
    "plt.tight_layout()\n",
    "plt.show()\n",
    "```\n",
    "\n",
    "```{figure} ../../img/xray/vinyl_calc.svg\n",
    "---\n",
    "name: xray_vinyl_calc\n",
    "scale: 6%\n",
    "---\n",
    "```\n",
    "\n",
    "(sec:xray_anal_amp)=\n",
    "## Looking at solution vectors\n",
    "\n",
    "The solution vectors, *i.e.* the eigenvectors for eigenstate calculations and the response vectors for CPP calculations, can be plotted in order to see involved occupied and virtual MOs. This can give some insight into the different features, but easily becomes cumbersome when the density of states is high, or when the final state consists of a mix of many contributions.\n",
    "\n",
    "### Eigenvectors\n",
    "\n",
    "From the ADC calculations there is the function `state.describe_amplitudes`, which print the amplitudes using one of three formats, controlled by `index_format`. The amplitudes are then printed when above a `tolerance` threshold. Splitting the returning string according to `state` and printing only the first two states, we get:\n",
    "\n",
    "```python\n",
    "print(\"From 'hf' format, which yield MO number:\")\n",
    "des_amp = adc_xas.describe_amplitudes(tolerance=0.1, index_format=\"hf\")\n",
    "split_line = \"\\n\\n\"\n",
    "tmp_amp = des_amp.split(split_line)\n",
    "print(tmp_amp[1])\n",
    "hf_exc1 = tmp_amp[1]\n",
    "print(tmp_amp[2])\n",
    "hf_exc2 = tmp_amp[1]\n",
    "print()\n",
    "\n",
    "print(\"\\nFirst transition from 'adcc' format, with indexing as used in adcc:\")\n",
    "des_amp = adc_xas.describe_amplitudes(tolerance=0.1, index_format=\"adcc\")\n",
    "split_line = \"\\n\\n\"\n",
    "tmp_amp = des_amp.split(split_line)\n",
    "print(tmp_amp[1])\n",
    "adcc_exc1 = tmp_amp[1]\n",
    "print()\n",
    "\n",
    "print(\"\\nFirst transition from 'homolumo' format, which use HOMO/LUMO/HOCO format:\")\n",
    "des_amp = adc_xas.describe_amplitudes(tolerance=0.1, index_format=\"homolumo\")\n",
    "split_line = \"\\n\\n\"\n",
    "tmp_amp = des_amp.split(split_line)\n",
    "homo_exc1 = tmp_amp[1]\n",
    "print(tmp_amp[1])\n",
    "```"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {
    "tags": [
     "remove-input"
    ]
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "From 'hf' format, which yield MO number:\n",
      "+---------------------------------------+\n",
      "| State   1 ,      10.63675 au          |\n",
      "+---------------------------------------+\n",
      "|   1     ->  12      a ->a      -0.601 |\n",
      "|   1     ->  22      a ->a      -0.105 |\n",
      "|  11   1 ->  12  12  ab->ab     +0.149 |\n",
      "+---------------------------------------+\n",
      "| State   1 ,      10.63675 au          |\n",
      "+---------------------------------------+\n",
      "|   1     ->  12      a ->a      -0.601 |\n",
      "|   1     ->  22      a ->a      -0.105 |\n",
      "|  11   1 ->  12  12  ab->ab     +0.149 |\n",
      "\n",
      "\n",
      "First transition from 'adcc' format, with indexing as used in adcc:\n",
      "+-------------------------------------------------------+\n",
      "| State   1 ,      10.63675 au,      289.4407 eV        |\n",
      "+-------------------------------------------------------+\n",
      "| (o2  1)         -> (v1  0)          a ->a      -0.601 |\n",
      "| (o2  1)         -> (v1 10)          a ->a      -0.105 |\n",
      "| (o1  8) (o2  1) -> (v1  0) (v1  0)  ab->ab     +0.149 |\n",
      "\n",
      "\n",
      "First transition from 'homolumo' format, which use HOMO/LUMO/HOCO format:\n",
      "+-------------------------------------------------------+\n",
      "| State   1 ,      10.63675 au,      289.4407 eV        |\n",
      "+-------------------------------------------------------+\n",
      "| HOCO -1         -> LUMO             a ->a      -0.601 |\n",
      "| HOCO -1         -> LUMO+10          a ->a      -0.105 |\n",
      "| HOMO    HOCO -1 -> LUMO    LUMO     ab->ab     +0.149 |\n"
     ]
    }
   ],
   "source": [
    "vinyl_mol_str = \"\"\"\n",
    " C     0.000000    -0.246412    -1.271068\n",
    " C     0.000000     0.457081    -0.154735\n",
    " F     0.000000    -0.119195     1.052878\n",
    " H     0.000000     0.272328    -2.210194\n",
    " H     0.000000    -1.319906    -1.249847\n",
    " H     0.000000     1.530323    -0.095954\n",
    "\"\"\"\n",
    "\n",
    "hf_exc1 = \"+---------------------------------------+\\n| State   1 ,      10.63675 au          |\\n+---------------------------------------+\\n|   1     ->  12      a ->a      -0.601 |\\n|   1     ->  22      a ->a      -0.105 |\\n|  11   1 ->  12  12  ab->ab     +0.149 |\"\n",
    "hf_exc2 = \"+---------------------------------------+\\n| State   1 ,      10.63675 au          |\\n+---------------------------------------+\\n|   1     ->  12      a ->a      -0.601 |\\n|   1     ->  22      a ->a      -0.105 |\\n|  11   1 ->  12  12  ab->ab     +0.149 |\"\n",
    "adcc_exc1 = \"+-------------------------------------------------------+\\n| State   1 ,      10.63675 au,      289.4407 eV        |\\n+-------------------------------------------------------+\\n| (o2  1)         -> (v1  0)          a ->a      -0.601 |\\n| (o2  1)         -> (v1 10)          a ->a      -0.105 |\\n| (o1  8) (o2  1) -> (v1  0) (v1  0)  ab->ab     +0.149 |\"\n",
    "homo_exc1 = \"+-------------------------------------------------------+\\n| State   1 ,      10.63675 au,      289.4407 eV        |\\n+-------------------------------------------------------+\\n| HOCO -1         -> LUMO             a ->a      -0.601 |\\n| HOCO -1         -> LUMO+10          a ->a      -0.105 |\\n| HOMO    HOCO -1 -> LUMO    LUMO     ab->ab     +0.149 |\"\n",
    "print(\"From 'hf' format, which yield MO number:\")\n",
    "print(hf_exc1)\n",
    "print(hf_exc2)\n",
    "print()\n",
    "print(\"\\nFirst transition from 'adcc' format, with indexing as used in adcc:\")\n",
    "print(adcc_exc1)\n",
    "print()\n",
    "print(\"\\nFirst transition from 'homolumo' format, which use HOMO/LUMO/HOCO format:\")\n",
    "print(homo_exc1)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "From this, it is clear that the two first transitions originate from two different occupied MOs, which correspond to the -**C**H$_2$ and -**C**HF atoms, respectively. The transitions are dominated by single-excitations to the LUMO and LUMO+1, with largest relaxation contributions coming from the HOMO.\n",
    "\n",
    "\n",
    "## Decomposition of spectra\n",
    "\n",
    "For a more visual illustration, we can plot the spectra as decomposed according to atomic contributions and polarization of absorbed radiation. \n",
    "\n",
    "\n",
    "### Atomic contributions\n",
    "\n",
    "Decomposing the spectra according to atomic contributions provides information on which core orbitals contribute to the various features. For vinylfluoride, this can be done by performing CVS calculations with tailored CVS spaces for each carbon site:\n",
    "\n",
    "```python\n",
    "# Construct structure and basis objects\n",
    "struct = gator.get_molecule(vinyl_mol_str)\n",
    "basis = gator.get_molecular_basis(struct, \"6-31G\")\n",
    "\n",
    "# Perform SCF calculation\n",
    "scf_gs = gator.run_scf(struct, basis)\n",
    "\n",
    "# Calculate the 6 lowest eigenstates with CVS restriction to MOs #2 and #3 (two carbons)\n",
    "adc_chf = gator.run_adc(\n",
    "    struct, basis, scf_gs, method=\"cvs-adc2x\", singlets=6, core_orbitals=[2]\n",
    ")\n",
    "adc_ch2 = gator.run_adc(\n",
    "    struct, basis, scf_gs, method=\"cvs-adc2x\", singlets=6, core_orbitals=[3]\n",
    ")\n",
    "```\n",
    "\n",
    "From which the contributions can be plotted:\n",
    "\n",
    "```python\n",
    "plt.figure(figsize=(7, 4))\n",
    "# Spectrum from -CHF\n",
    "x1, y1 = au2ev * adc_chf.excitation_energy, adc_chf.oscillator_strength\n",
    "x1i, y1i = lorentzian(x1, y1, 280, 300, 0.01, 0.8)\n",
    "\n",
    "# Spectrum from -CH2\n",
    "x2, y2 = au2ev * adc_ch2.excitation_energy, adc_ch2.oscillator_strength\n",
    "x2i, y2i = lorentzian(x2, y2, 280, 300, 0.01, 0.8)\n",
    "\n",
    "# Plotting line spectra for full, area spectra for contributions\n",
    "plt.plot(x2i, y1i + y2i, \"k-\", linewidth=2.0)\n",
    "plt.fill_between(x2i, 0, y1i, alpha=0.6)\n",
    "plt.fill_between(x2i, y1i, y1i + y2i, alpha=0.6)\n",
    "\n",
    "plt.legend((\"Full\", \"-CHF\", \"-CH2\"), loc=\"upper left\")\n",
    "plt.xlim(284, 296)\n",
    "plt.tight_layout()\n",
    "plt.show()\n",
    "```\n",
    "\n",
    "```{figure} ../../img/xray/adc_decomp.svg\n",
    "---\n",
    "name: xray_adc_decomp\n",
    "scale: 6%\n",
    "---\n",
    "```\n",
    "\n",
    "We here clearly see the chemical shift due to fluorine-substitution.\n",
    "\n",
    "This manner of decomposing the atomic contributions should be taken with some care, as it involves explicit calculation for each atomic site in turn, which:\n",
    "\n",
    "1. Can require a larger number of calculations, and care should be taken so that the studied region does contain all states of one atomic site, but not another.\n",
    "\n",
    "2. This approach does not strictly fit with the CVS philosophy, which uses the large separation in energy and spatial extent as a basis for separation. Energy separation can now be small, and for some systems degeneracies and near-degeneracies may lead to delocalized core orbitals (see [here](sec:xray_topics_chloc) and [here](sec:xray_topics_tailcvs)). If so, all delocalized sites must be included in the same CVS space.\n",
    "\n",
    "Nevertheless, this approach has some advantages in terms of ease of analysis, as well as exhibiting lower computational cost for each calculation (albeit here requiring two calculations instead of just one).\n",
    "\n",
    "\n",
    "An alternative to this is to perform the decomposition on the full solution vectors, as will here be done using CPP-DFT response vectors:\n",
    "\n",
    "\n",
    "```python\n",
    "# Number of occupied and unoccupied (alpha) MOs\n",
    "nocc = molecule.number_of_alpha_electrons()\n",
    "nvirt = scf_drv.mol_orbs.number_mos() - nocc\n",
    "\n",
    "# Extract solution vectors\n",
    "solution_vecs = cpp_prop.get_property(\"solutions\")\n",
    "x_solution, y_solution, z_solution = [], [], []\n",
    "for w in freqs:\n",
    "    x_solution.append(cpp_prop.get_full_solution_vector(solution_vecs[(\"x\", w)]))\n",
    "    y_solution.append(cpp_prop.get_full_solution_vector(solution_vecs[(\"y\", w)]))\n",
    "    z_solution.append(cpp_prop.get_full_solution_vector(solution_vecs[(\"z\", w)]))\n",
    "    \n",
    "x_solution, y_solution, z_solution = (\n",
    "    np.array(x_solution),\n",
    "    np.array(y_solution),\n",
    "    np.array(z_solution),\n",
    ")\n",
    "\n",
    "# Extract polarization-resolved response\n",
    "sX, sY, sZ = [], [], []\n",
    "for w in freqs:\n",
    "    axx = -cpp_prop.rsp_property[\"response_functions\"][(\"x\", \"x\", w)].imag\n",
    "    ayy = -cpp_prop.rsp_property[\"response_functions\"][(\"y\", \"y\", w)].imag\n",
    "    azz = -cpp_prop.rsp_property[\"response_functions\"][(\"z\", \"z\", w)].imag\n",
    "    sX.append(4.0 / 3.0 * np.pi * w * axx / 137.035999)\n",
    "    sY.append(4.0 / 3.0 * np.pi * w * ayy / 137.035999)\n",
    "    sZ.append(4.0 / 3.0 * np.pi * w * azz / 137.035999)\n",
    "\n",
    "# Create empty object of dimension nocc (to add spectrum contributions from each occupied orbital)\n",
    "spec_comp = []\n",
    "for occ in np.arange(0, nocc):\n",
    "    spec_comp.append([])\n",
    "\n",
    "# Appending spec_comp object with decomposed response\n",
    "for i in np.arange(len(freqs)):\n",
    "    x_tmp = x_solution[i]\n",
    "    y_tmp = y_solution[i]\n",
    "    z_tmp = z_solution[i]\n",
    "    for occ in np.arange(nocc):\n",
    "        comp_X = np.sum(\n",
    "            np.abs(np.imag(x_tmp[occ * nvirt : (occ + 1) * nvirt]))\n",
    "        ) / np.sum(np.abs(np.imag(x_tmp[:])))\n",
    "        comp_Y = np.sum(\n",
    "            np.abs(np.imag(y_tmp[occ * nvirt : (occ + 1) * nvirt]))\n",
    "        ) / np.sum(np.abs(np.imag(y_tmp[:])))\n",
    "        comp_Z = np.sum(\n",
    "            np.abs(np.imag(z_tmp[occ * nvirt : (occ + 1) * nvirt]))\n",
    "        ) / np.sum(np.abs(np.imag(z_tmp[:])))\n",
    "        spec_comp[occ].append((comp_X * sX[i] + comp_Y * sY[i] + comp_Z * sZ[i]))\n",
    "\n",
    "# Extract contribution of two carbon atoms, as well as the full spectrum\n",
    "cont_chf = np.array(spec_comp[1])\n",
    "cont_ch2 = np.array(spec_comp[2])\n",
    "cont_full = sigma\n",
    "\n",
    "plt.figure(figsize=(7, 4))\n",
    "# Create splined spectra for each component\n",
    "x = np.arange(min(au2ev * freqs), max(au2ev * freqs), 0.01)\n",
    "ychf = interp1d(au2ev * freqs, cont_chf, kind=\"cubic\")\n",
    "ych2 = interp1d(au2ev * freqs, cont_ch2, kind=\"cubic\")\n",
    "yfull = interp1d(au2ev * freqs, cont_full, kind=\"cubic\")\n",
    "# Spectrum from the non-carbon atoms\n",
    "yrest = yfull(x) - ychf(x) - ych2(x)\n",
    "# Plot full spectrum\n",
    "plt.plot(x, yfull(x), \"k-\", linewidth=2)\n",
    "# Plot components\n",
    "plt.fill_between(x, yrest, yrest + ychf(x), alpha=0.6)\n",
    "plt.fill_between(x, yrest + ychf(x), yrest + ych2(x) + ychf(x), alpha=0.6)\n",
    "plt.fill_between(x, 0, yrest, alpha=0.6)\n",
    "\n",
    "plt.legend((\"Total\", \"-CHF\", r\"-CH$_2$\", \"Rest\"))\n",
    "plt.xlim((272, 284))\n",
    "plt.tight_layout()\n",
    "plt.show()\n",
    "```\n",
    "\n",
    "```{figure} ../../img/xray/dft_decomp.svg\n",
    "---\n",
    "name: xray_dft_decomp\n",
    "scale: 6%\n",
    "---\n",
    "```\n",
    "\n",
    "Where we again clearly see the chemical shift and different contributions.\n",
    "\n",
    "\n",
    "### Polarization dependence\n",
    "\n",
    "For structured molecular samples, the local orientation of the molecules can be studied by looking at the polarization-resolved spectra. This will not work for the gas phase, where the isotropic signal is observed, but it can provide very useful information on, *e.g.*, the orientation and structuring of self-assembly monolayers. \n",
    "\n",
    "Resolving the polarization-dependence from ADC and CPP-DFT calculations is straightforward:\n",
    "\n",
    "\n",
    "```python\n",
    "# Extract excitation energies\n",
    "adc_freq = adc_xas.excitation_energy\n",
    "\n",
    "# Extract Cartesian components from the ADC state (here skipping the prefactors)\n",
    "adc_sX, adc_sY, adc_sZ = [], [], []\n",
    "for i in np.arange(len(adc_freq)):\n",
    "    tmp_tms = adc_xas.transition_dipole_moment[i]\n",
    "    adc_sX.append(adc_freq[i] * tmp_tms[0] ** 2)\n",
    "    adc_sY.append(adc_freq[i] * tmp_tms[1] ** 2)\n",
    "    adc_sZ.append(adc_freq[i] * tmp_tms[2] ** 2)\n",
    "\n",
    "plt.figure(figsize=(7, 4))\n",
    "# Resolved spectrum region\n",
    "xmin, xmax = 284, 296\n",
    "\n",
    "# Broadened spectra for each component\n",
    "x, y = au2ev * adc_freq, adc_sX\n",
    "xX, yX = lorentzian(x, y, xmin, xmax, 0.01, 0.8)\n",
    "x, y = au2ev * adc_freq, adc_sY\n",
    "xY, yY = lorentzian(x, y, xmin, xmax, 0.01, 0.8)\n",
    "x, y = au2ev * adc_freq, adc_sZ\n",
    "xZ, yZ = lorentzian(x, y, xmin, xmax, 0.01, 0.8)\n",
    "\n",
    "# Plot total spectrum (line) and Cartesian contributions (area)\n",
    "plt.plot(xX, yX + yY + yZ, \"k-\", linewidth=2)\n",
    "plt.fill_between(xX, 0, yX, alpha=0.6)\n",
    "plt.fill_between(xX, yX, yX + yY, alpha=0.6)\n",
    "plt.fill_between(xX, yX + yY, yX + yY + yZ, alpha=0.6)\n",
    "\n",
    "plt.legend((\"Isotropic\", \"x-pol.\", \"y-pol.\", \"z-pol.\"))\n",
    "plt.xlim((xmin, xmax))\n",
    "plt.tight_layout()\n",
    "plt.show()\n",
    "```\n",
    "\n",
    "```{figure} ../../img/xray/adc_pol.svg\n",
    "---\n",
    "name: xray_adc_pol\n",
    "scale: 6%\n",
    "---\n",
    "```\n",
    "\n",
    "\n",
    "```python\n",
    "plt.figure(figsize=(7, 4))\n",
    "# Extract frequencies\n",
    "x = np.arange(min(au2ev * freqs), max(au2ev * freqs), 0.01)\n",
    "\n",
    "# Spline spectra for each Cartesian contribution\n",
    "yX = interp1d(au2ev * freqs, sX, kind=\"cubic\")\n",
    "yY = interp1d(au2ev * freqs, sY, kind=\"cubic\")\n",
    "yZ = interp1d(au2ev * freqs, sZ, kind=\"cubic\")\n",
    "\n",
    "# Plot total spectrum (line) and Cartesian contributions (area)\n",
    "plt.plot(x, yX(x) + yY(x) + yZ(x), \"k-\", linewidth=2)\n",
    "plt.fill_between(x, 0, yX(x), alpha=0.6)\n",
    "plt.fill_between(x, yX(x), yX(x) + yY(x), alpha=0.6)\n",
    "plt.fill_between(x, yX(x) + yY(x), yX(x) + yY(x) + yZ(x), alpha=0.6)\n",
    "\n",
    "plt.legend((\"Isotropic\", \"x-pol.\", \"y-pol.\", \"z-pol.\"))\n",
    "plt.xlim((272, 284))\n",
    "plt.tight_layout()\n",
    "plt.show()\n",
    "```\n",
    "\n",
    "\n",
    "```{figure} ../../img/xray/dft_pol.svg\n",
    "---\n",
    "name: xray_dft_pol\n",
    "scale: 6%\n",
    "---\n",
    "```\n",
    "\n",
    "We see that the first two transitions are out-of-plane, correspond to the two strong $1s \\rightarrow \\pi^{\\ast}$ excitations. As such, if the molecule is rotated in space, these two lines would provide very good probes on the angle with respect to the out-of-plane direction."
   ]
  }
 ],
 "metadata": {
  "celltoolbar": "Tags",
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.4"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
